# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: 
"""
import hashlib
from datetime import datetime, timezone
from typing import Dict, Any, Optional, List, Union
from dataclasses import asdict


def remove_indent(s: Optional[str]) -> Optional[str]:
    """
    Remove the indent from a string.

    Args:
        s (str): String to remove indent from

    Returns:
        str: String with indent removed
    """
    if s is not None and isinstance(s, str):
        return "\n".join([line.strip() for line in s.split("\n")])
    return None


def merge_dictionaries(a: Dict[str, Any], b: Dict[str, Any]) -> None:
    """
    Recursively merges two dictionaries.
    If there are conflicting keys, values from 'b' will take precedence.

    @params:
    a (Dict[str, Any]): The first dictionary to be merged.
    b (Dict[str, Any]): The second dictionary, whose values will take precedence.

    Returns:
    None: The function modifies the first dictionary in place.
    """
    for key in b:
        if key in a and isinstance(a[key], dict) and isinstance(b[key], dict):
            merge_dictionaries(a[key], b[key])
        else:
            a[key] = b[key]


def current_datetime() -> datetime:
    return datetime.now()


def current_datetime_utc() -> datetime:
    return datetime.now(timezone.utc)


def current_datetime_utc_str() -> str:
    return current_datetime_utc().strftime("%Y-%m-%dT%H:%M:%S")


def calculate_sha256(file):
    sha256 = hashlib.sha256()
    # Read the file in chunks to efficiently handle large files
    for chunk in iter(lambda: file.read(8192), b""):
        sha256.update(chunk)
    return sha256.hexdigest()


def calculate_sha256_string(string):
    # Create a new SHA-256 hash object
    sha256_hash = hashlib.sha256()
    # Update the hash object with the bytes of the input string
    sha256_hash.update(string.encode("utf-8"))
    # Get the hexadecimal representation of the hash
    hashed_string = sha256_hash.hexdigest()
    return hashed_string


def literal_similarity(text1, text2):
    """
    判断两个文本的字面相似度

    参数:
    text1: 第一个文本字符串
    text2: 第二个文本字符串

    返回值:
    相似度分数，介于0和1之间，其中1表示完全相同
    """
    from difflib import SequenceMatcher
    matcher = SequenceMatcher(None, text1, text2)
    return matcher.ratio()


def dataclass_to_dict(dataclass_object, exclude: Optional[set[str]] = None, exclude_none: bool = False) -> dict:
    final_dict = asdict(dataclass_object)
    if exclude:
        for key in exclude:
            final_dict.pop(key)
    if exclude_none:
        final_dict = {k: v for k, v in final_dict.items() if v is not None}
    return final_dict



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: 
"""
import sys

from loguru import logger as _logger

from agentica.config import AGENTICA_LOG_FILE, AGENTICA_LOG_LEVEL


def get_logger(log_level: str = "INFO", log_file: str = None):
    """Get logger instance."""
    _logger.remove()
    _logger.add(
        sink=sys.stdout,
        level=log_level,
        enqueue=True,  # 异步输出日志
        backtrace=False,  # 设置为 True 以打印回溯
        diagnose=False,  # 设置为 True 以自动显示变量
    )
    if log_file:
        _logger.add(
            sink=log_file,
            level=log_level,
            rotation="1 week",
            retention="1 month",
            enqueue=True,
            backtrace=False,
            diagnose=False,
        )
    return _logger


logger = get_logger(log_level=AGENTICA_LOG_LEVEL, log_file=AGENTICA_LOG_FILE)


def set_log_level_to_debug():
    logger.remove()  # 移除之前的日志处理器
    logger.add(
        sink=sys.stdout,
        level="DEBUG",
        enqueue=True,
        backtrace=True,
        diagnose=True,
    )
    if AGENTICA_LOG_FILE:
        logger.add(
            sink=AGENTICA_LOG_FILE,
            level="DEBUG",
            enqueue=True,
            backtrace=True,
            diagnose=True,
        )


def set_log_level_to_info():
    logger.remove()  # 移除之前的日志处理器
    logger.add(
        sink=sys.stdout,
        level="INFO",
        enqueue=True,
        backtrace=False,
        diagnose=False,
    )
    if AGENTICA_LOG_FILE:
        logger.add(
            sink=AGENTICA_LOG_FILE,
            level="INFO",
            enqueue=True,
            backtrace=False,
            diagnose=False,
        )


def print_llm_stream(msg):
    print(msg, end="", flush=True)



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: command line interface for agentica
"""
import argparse
import os
import sys
import time
from rich.console import Console
from rich.text import Text
import importlib
from typing import List, Optional
from agentica import Agent, OpenAIChat, Moonshot, AzureOpenAIChat, Yi, ZhipuAI, DeepSeek, PythonAgent


console = Console()

# Color constants
class TermColor:
    BRIGHT_CYAN = "\033[96m"
    BRIGHT_GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"
    WHITE = "\033[97m"
    RESET = "\033[0m"

# Tool mapping dictionary - maps tool names to their import paths
TOOL_MAP = {
    'airflow': 'agentica.tools.airflow_tool.AirflowTool',
    'analyze_image': 'agentica.tools.analyze_image_tool.AnalyzeImageTool',
    'arxiv': 'agentica.tools.arxiv_tool.ArxivTool',
    'baidu_search': 'agentica.tools.baidu_search_tool.BaiduSearchTool',
    'calculator': 'agentica.tools.calculator_tool.CalculatorTool',
    'code': 'agentica.tools.code_tool.CodeTool',
    'cogview': 'agentica.tools.cogview_tool.CogViewTool',
    'cogvideo': 'agentica.tools.cogvideo_tool.CogVideoTool',
    'dalle': 'agentica.tools.dalle_tool.DalleTool',
    'duckduckgo': 'agentica.tools.duckduckgo_tool.DuckDuckGoTool',
    'edit': 'agentica.tools.edit_tool.EditTool',
    'file': 'agentica.tools.file_tool.FileTool',
    'hackernews': 'agentica.tools.hackernews_tool.HackerNewsTool',
    'jina': 'agentica.tools.jina_tool.JinaTool',
    'mcp': 'agentica.tools.mcp_tool.MCPTool',
    'newspaper': 'agentica.tools.newspaper_tool.NewspaperTool',
    'ocr': 'agentica.tools.ocr_tool.OcrTool',
    'run_python_code': 'agentica.tools.run_python_code_tool.RunPythonCodeTool',
    'search_exa': 'agentica.tools.search_exa_tool.SearchExaTool',
    'search_serper': 'agentica.tools.search_serper_tool.SearchSerperTool',
    'shell': 'agentica.tools.shell_tool.ShellTool',
    'string': 'agentica.tools.string_tool.StringTool',
    'text_analysis': 'agentica.tools.text_analysis_tool.TextAnalysisTool',
    'url_crawler': 'agentica.tools.url_crawler_tool.UrlCrawlerTool',
    'volc_tts': 'agentica.tools.volc_tts_tool.VolcTtsTool',
    'weather': 'agentica.tools.weather_tool.WeatherTool',
    'web_search_pro': 'agentica.tools.web_search_pro_tool.WebSearchProTool',
    'wikipedia': 'agentica.tools.wikipedia_tool.WikipediaTool',
    'workspace': 'agentica.tools.workspace_tool.WorkspaceTool',
    'yfinance': 'agentica.tools.yfinance_tool.YFinanceTool',
}


def parse_args():
    parser = argparse.ArgumentParser(description='CLI for agentica')
    parser.add_argument('--query', type=str, help='Question to ask the LLM', default=None)
    parser.add_argument('--model_provider', type=str,
                        choices=['openai', 'azure', 'moonshot', 'zhipuai', 'deepseek', 'yi'],
                        help='LLM model provider', default='openai')
    parser.add_argument('--model_name', type=str,
                        help='LLM model name to use, can be gpt-4o/glm-4-flash/deepseek-chat/yi-lightning/...',
                        default='gpt-4o-mini')
    parser.add_argument('--api_base', type=str, help='API base URL for the LLM')
    parser.add_argument('--api_key', type=str, help='API key for the LLM')
    parser.add_argument('--max_tokens', type=int, help='Maximum number of tokens for the LLM')
    parser.add_argument('--temperature', type=float, help='Temperature for the LLM')
    parser.add_argument('--verbose', type=int, help='enable verbose mode', default=0)
    parser.add_argument('--tools', nargs='*',
                        choices=list(TOOL_MAP.keys()),
                        help='Tools to enable')
    return parser.parse_args()


def get_model(model_provider, model_name, api_base=None, api_key=None, max_tokens=None, temperature=None):
    params = {"id": model_name}
    if api_base is not None:
        params["api_base"] = api_base
    if api_key is not None:
        params["api_key"] = api_key
    if max_tokens is not None:
        params["max_tokens"] = max_tokens
    if temperature is not None:
        params["temperature"] = temperature
    if model_provider == 'openai':
        model = OpenAIChat(**params)
    elif model_provider == 'azure':
        model = AzureOpenAIChat(**params)
    elif model_provider == 'moonshot':
        model = Moonshot(**params)
    elif model_provider == 'zhipuai':
        model = ZhipuAI(**params)
    elif model_provider == 'deepseek':
        model = DeepSeek(**params)
    elif model_provider == 'yi':
        model = Yi(**params)
    else:
        raise ValueError(f"Unsupported model: {model_name}")
    return model


def configure_tools(tool_names: Optional[List[str]] = None) -> List:
    """
    Configure and instantiate tools based on their names.
    Uses dynamic import to load tools only when requested.

    Args:
        tool_names: List of tool names to enable. Must be keys in TOOL_MAP.

    Returns:
        List of instantiated tool objects.
    """
    if not tool_names:
        return []

    tools = []
    for name in tool_names:
        if name not in TOOL_MAP:
            console.print(f"[yellow]Warning: Tool '{name}' not recognized. Skipping.[/yellow]")
            continue

        try:
            # Dynamically import and instantiate the tool
            module_path, class_name = TOOL_MAP[name].rsplit('.', 1)
            module = importlib.import_module(module_path)
            tool_class = getattr(module, class_name)
            tool_instance = tool_class()
            tools.append(tool_instance)
            console.print(f"[green]Successfully loaded tool: {name}[/green]")
        except ImportError as e:
            console.print(f"[red]Error: Could not import tool '{name}'. Missing dependencies? {str(e)}[/red]")
        except Exception as e:
            console.print(f"[red]Error: Failed to initialize tool '{name}': {str(e)}[/red]")

    return tools


def print_header(model_provider: str, model_name: str, tools: Optional[List[str]] = None):
    """Print the application header with version and model information"""
    header_color = TermColor.BRIGHT_CYAN
    accent_color = TermColor.BRIGHT_GREEN
    reset = TermColor.RESET

    # Create a boxed header
    box_width = 60
    border_top = f"{header_color}╭{'─' * (box_width - 1)}╮{reset}"
    border_bottom = f"{header_color}╰{'─' * (box_width - 1)}╯{reset}"

    # Print header
    print(border_top)

    # App name and version
    app_name = f"{header_color}│{reset} {accent_color}Agentica CLI{reset} - Interactive AI Assistant"
    padding = box_width - len(app_name) + len(header_color) + len(reset) + len(accent_color) + len(reset)
    print(f"{app_name}{' ' * padding}{header_color}│{reset}")

    # Model info
    model_info = f"{header_color}│{reset} Model: {accent_color}{model_provider}/{model_name}{reset}"
    padding = box_width - len(model_info) + len(header_color) + len(reset) + len(accent_color) + len(reset)
    print(f"{model_info}{' ' * padding}{header_color}│{reset}")

    # Tools info
    if tools:
        tools_str = ", ".join(tools)
        tools_line = f"{header_color}│{reset} Tools: {accent_color}{tools_str}{reset}"
        if len(tools_line) - len(header_color) - len(reset) - len(accent_color) - len(reset) > box_width - 4:
            tools_str = tools_str[:(box_width - 24)] + "..."
            tools_line = f"{header_color}│{reset} Tools: {accent_color}{tools_str}{reset}"
        padding = box_width - len(tools_line) + len(header_color) + len(reset) + len(accent_color) + len(reset)
        print(f"{tools_line}{' ' * padding}{header_color}│{reset}")

    # Working directory
    cwd = os.getcwd()
    home = os.path.expanduser("~")
    if cwd.startswith(home):
        cwd = "~" + cwd[len(home):]
    cwd_line = f"{header_color}│{reset} Working Directory: {cwd}"
    if len(cwd_line) - len(header_color) - len(reset) > box_width - 4:
        display_cwd = "..." + cwd[-(box_width - 24):]
        cwd_line = f"{header_color}│{reset} Working Directory: {display_cwd}"
    padding = box_width - len(cwd_line) + len(header_color) + len(reset)
    print(f"{cwd_line}{' ' * padding}{header_color}│{reset}")

    # Input instructions
    input_line = f"{header_color}│{reset} Input with {accent_color}>{reset} prompt, press {accent_color}Enter twice{reset} to submit"
    padding = box_width - len(input_line) + len(header_color) + len(reset) + 2 * len(accent_color) + 2 * len(reset)
    print(f"{input_line}{' ' * padding}{header_color}│{reset}")

    print(border_bottom)
    print()


def run_interactive(agent):
    """Run the interactive CLI with multi-line input support"""
    thinking_start_time = 0
    loading = False

    def show_thinking_indicator():
        """Show a thinking indicator with elapsed time while waiting for a response"""
        if loading:
            elapsed = int(time.time() - thinking_start_time)
            sys.stdout.write(f"\rThinking... ({elapsed}s)")
            sys.stdout.flush()

    def clear_thinking_indicator():
        """Clear the thinking indicator from the terminal"""
        sys.stdout.write("\r" + " " * 30 + "\r")  # Clear the line
        sys.stdout.flush()

    def get_user_input():
        """Get multi-line user input. Input ends when user enters a single empty line."""
        try:
            console.print(Text("> ", style="green"), end="")
            lines = []
            while True:
                try:
                    line = input()
                    if not line.strip():  # Empty line terminates input
                        break
                    lines.append(line)
                except EOFError:
                    return None

            # If no valid input, return None
            if not lines or not any(line.strip() for line in lines):
                return None

            user_input = '\n'.join(lines)
            user_input = user_input.strip()

            # Check if user wants to exit
            if user_input.lower() in ['exit', 'quit', '\\q']:
                return None

            return user_input
        except KeyboardInterrupt:
            print("\nOperation interrupted.")
            return None

    while True:
        try:
            query = get_user_input()
            if query is None:
                break

            # Set loading state and start thinking indicator
            loading = True
            thinking_start_time = time.time()

            # Start a timer to update the thinking indicator
            import threading
            def update_indicator():
                while loading:
                    show_thinking_indicator()
                    time.sleep(1)  # Update every second

            indicator_thread = threading.Thread(target=update_indicator)
            indicator_thread.daemon = True
            indicator_thread.start()

            # Get response from agent
            response = agent.run(query, stream=True)
            first_chunk = True

            # Print response
            for chunk in response:
                if first_chunk:
                    loading = False  # Stop the thinking indicator
                    indicator_thread.join(timeout=0.1)  # Wait for indicator thread to finish
                    clear_thinking_indicator()  # Clear the indicator
                    first_chunk = False
                console.print(chunk.content, end="")
            console.print("\n")

        except KeyboardInterrupt:
            loading = False
            print("\nOperation interrupted.")
            continue
        except Exception as e:
            loading = False
            console.print(f"[red]Error: {str(e)}[/red]")
            continue

    console.print("\nThank you for using Agentica CLI. Goodbye!")


def main():
    args = parse_args()
    model = get_model(args.model_provider, args.model_name, api_base=args.api_base,
                      api_key=args.api_key, max_tokens=args.max_tokens)
    tools = configure_tools(args.tools) if args.tools else None
    debug_mode = args.verbose > 0
    if args.tools and 'python' in args.tools:
        agent = PythonAgent(model=model, add_datetime_to_instructions=True, add_history_to_messages=True,
                            tools=tools, debug_mode=debug_mode)
    else:
        agent = Agent(model=model, add_datetime_to_instructions=True, add_history_to_messages=True,
                      tools=tools, debug_mode=debug_mode)

    if args.query:
        response = agent.run(args.query, stream=True)
        for chunk in response:
            console.print(chunk.content, end="")
    else:
        print_header(args.model_provider, args.model_name, args.tools)
        run_interactive(agent)


if __name__ == "__main__":
    main()



import collections.abc

from types import GeneratorType
from typing import List, Iterator, Optional, Dict, Any, Callable, Union, Sequence

from pydantic import BaseModel, ConfigDict, Field, field_validator, ValidationInfo

from agentica.model.message import Message
from agentica.model.response import ModelResponse, ModelResponseEvent
from agentica.tools.base import Tool, Toolkit, Function, FunctionCall, ToolCallException
from agentica.utils.log import logger
from agentica.utils.timer import Timer


class Model(BaseModel):
    # ID of the model to use.
    id: str = Field(..., alias="model")
    # Name for this Model. This is not sent to the Model API.
    name: Optional[str] = None
    # Provider for this Model. This is not sent to the Model API.
    provider: Optional[str] = Field(None, validate_default=True)
    # Metrics collected for this Model. This is not sent to the Model API.
    metrics: Dict[str, Any] = Field(default_factory=dict)
    response_format: Optional[Any] = None

    # A list of tools provided to the Model.
    # Tools are functions the model may generate JSON inputs for.
    # If you provide a dict, it is not called by the model.
    # Always add tools using the add_tool() method.
    tools: Optional[List[Union[Tool, Dict]]] = None
    # Controls which (if any) function is called by the model.
    # "none" means the model will not call a function and instead generates a message.
    # "auto" means the model can pick between generating a message or calling a function.
    # Specifying a particular function via {"type: "function", "function": {"name": "my_function"}}
    #   forces the model to call that function.
    # "none" is the default when no functions are present. "auto" is the default if functions are present.
    tool_choice: Optional[Union[str, Dict[str, Any]]] = None
    # If True, runs the tool before sending back the response content.
    run_tools: bool = True
    # If True, shows function calls in the response.
    show_tool_calls: Optional[bool] = None
    # Maximum number of tool calls allowed.
    tool_call_limit: Optional[int] = None

    # -*- Functions available to the Model to call -*-
    # Functions extracted from the tools.
    # Note: These are not sent to the Model API and are only used for execution + deduplication.
    functions: Optional[Dict[str, Function]] = None
    # Function call stack.
    function_call_stack: Optional[List[FunctionCall]] = None

    # System prompt from the model added to the Agent.
    system_prompt: Optional[str] = None
    # Instructions from the model added to the Agent.
    instructions: Optional[List[str]] = None

    # Session ID of the calling Agent or Workflow.
    session_id: Optional[str] = None
    # Whether to use the structured outputs with this Model.
    structured_outputs: Optional[bool] = None
    # Whether the Model supports structured outputs.
    supports_structured_outputs: bool = False

    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    @field_validator("provider", mode="before")
    def set_provider(cls, v: Optional[str], info: ValidationInfo) -> str:
        model_name = info.data.get("name")
        model_id = info.data.get("id")
        return v or f"{model_name} ({model_id})"

    @property
    def request_kwargs(self) -> Dict[str, Any]:
        raise NotImplementedError

    def to_dict(self) -> Dict[str, Any]:
        _dict = self.model_dump(include={"name", "id", "provider", "metrics"})
        if self.functions:
            _dict["functions"] = {k: v.to_dict() for k, v in self.functions.items()}
            _dict["tool_call_limit"] = self.tool_call_limit
        return _dict

    def invoke(self, *args, **kwargs) -> Any:
        raise NotImplementedError

    async def ainvoke(self, *args, **kwargs) -> Any:
        raise NotImplementedError

    def invoke_stream(self, *args, **kwargs) -> Iterator[Any]:
        raise NotImplementedError

    async def ainvoke_stream(self, *args, **kwargs) -> Any:
        raise NotImplementedError

    def response(self, messages: List[Message]) -> ModelResponse:
        raise NotImplementedError

    async def aresponse(self, messages: List[Message]) -> ModelResponse:
        raise NotImplementedError

    def response_stream(self, messages: List[Message]) -> Iterator[ModelResponse]:
        raise NotImplementedError

    async def aresponse_stream(self, messages: List[Message]) -> Any:
        raise NotImplementedError

    def _log_messages(self, messages: List[Message]) -> None:
        """
        Log messages for debugging.
        """
        for m in messages:
            m.log()

    def get_tools_for_api(self) -> Optional[List[Dict[str, Any]]]:
        if self.tools is None:
            return None

        tools_for_api = []
        for tool in self.tools:
            if isinstance(tool, Tool):
                tools_for_api.append(tool.to_dict())
            elif isinstance(tool, Dict):
                tools_for_api.append(tool)
        return tools_for_api

    def add_tool(
            self, tool: Union[Tool, Toolkit, Callable, Dict, Function], strict: bool = False,
            agent: Optional[Any] = None
    ) -> None:
        if self.tools is None:
            self.tools = []

        # If the tool is a Tool or Dict, add it directly to the Model
        if isinstance(tool, Tool) or isinstance(tool, Dict):
            if tool not in self.tools:
                self.tools.append(tool)
                logger.debug(f"Added tool {tool} to model.")

        # If the tool is a Callable or Toolkit, process and add to the Model
        elif callable(tool) or isinstance(tool, Toolkit) or isinstance(tool, Function):
            if self.functions is None:
                self.functions = {}

            if isinstance(tool, Toolkit):
                # For each function in the toolkit, process entrypoint and add to self.tools
                for name, func in tool.functions.items():
                    # If the function does not exist in self.functions, add to self.tools
                    if name not in self.functions:
                        func._agent = agent
                        func.process_entrypoint(strict=strict)
                        if strict and self.supports_structured_outputs:
                            func.strict = True
                        self.functions[name] = func
                        self.tools.append({"type": "function", "function": func.to_dict()})
                        logger.debug(f"Function {name} from {tool.name} added to model.")

            elif isinstance(tool, Function):
                if tool.name not in self.functions:
                    tool._agent = agent
                    tool.process_entrypoint(strict=strict)
                    if strict and self.supports_structured_outputs:
                        tool.strict = True
                    self.functions[tool.name] = tool
                    self.tools.append({"type": "function", "function": tool.to_dict()})
                    logger.debug(f"Function {tool.name} added to model.")

            elif callable(tool):
                try:
                    function_name = tool.__name__
                    if function_name not in self.functions:
                        func = Function.from_callable(tool, strict=strict)
                        func._agent = agent
                        if strict and self.supports_structured_outputs:
                            func.strict = True
                        self.functions[func.name] = func
                        self.tools.append({"type": "function", "function": func.to_dict()})
                        logger.debug(f"Function {func.name} added to model.")
                except Exception as e:
                    logger.warning(f"Could not add function {tool}: {e}")

    def deactivate_function_calls(self) -> None:
        # Deactivate tool calls by setting future tool calls to "none"
        # This is triggered when the function call limit is reached.
        self.tool_choice = "none"

    def run_function_calls(
            self, function_calls: List[FunctionCall], function_call_results: List[Message], tool_role: str = "tool"
    ) -> Iterator[ModelResponse]:
        for function_call in function_calls:
            if self.function_call_stack is None:
                self.function_call_stack = []

            # -*- Start function call
            function_call_timer = Timer()
            function_call_timer.start()
            yield ModelResponse(
                content=function_call.get_call_str(),
                tool_call={
                    "role": tool_role,
                    "tool_call_id": function_call.call_id,
                    "tool_name": function_call.function.name,
                    "tool_args": function_call.arguments,
                },
                event=ModelResponseEvent.tool_call_started.value,
            )

            # Track if the function call was successful
            function_call_success = False
            # If True, stop execution after this function call
            stop_execution_after_tool_call = False
            # Additional messages from the function call that will be added to the function call results
            additional_messages_from_function_call = []

            # -*- Run function call
            try:
                function_call_success = function_call.execute()
            except ToolCallException as tce:
                if tce.user_message is not None:
                    if isinstance(tce.user_message, str):
                        additional_messages_from_function_call.append(Message(role="user", content=tce.user_message))
                    else:
                        additional_messages_from_function_call.append(tce.user_message)
                if tce.agent_message is not None:
                    if isinstance(tce.agent_message, str):
                        additional_messages_from_function_call.append(
                            Message(role="assistant", content=tce.agent_message)
                        )
                    else:
                        additional_messages_from_function_call.append(tce.agent_message)
                if tce.messages is not None and len(tce.messages) > 0:
                    for m in tce.messages:
                        if isinstance(m, Message):
                            additional_messages_from_function_call.append(m)
                        elif isinstance(m, dict):
                            try:
                                additional_messages_from_function_call.append(Message(**m))
                            except Exception as e:
                                logger.warning(f"Failed to convert dict to Message: {e}")
                if tce.stop_execution:
                    stop_execution_after_tool_call = True
                    if len(additional_messages_from_function_call) > 0:
                        for m in additional_messages_from_function_call:
                            m.stop_after_tool_call = True

            function_call_output: Optional[Union[List[Any], str]] = ""
            if isinstance(function_call.result, (GeneratorType, collections.abc.Iterator)):
                for item in function_call.result:
                    function_call_output += item
                    if function_call.function.show_result:
                        yield ModelResponse(content=item)
            else:
                function_call_output = function_call.result
                if function_call.function.show_result:
                    yield ModelResponse(content=function_call_output)

            # -*- Stop function call timer
            function_call_timer.stop()

            # -*- Create function call result message
            function_call_result = Message(
                role=tool_role,
                content=function_call_output if function_call_success else function_call.error,
                tool_call_id=function_call.call_id,
                tool_name=function_call.function.name,
                tool_args=function_call.arguments,
                tool_call_error=not function_call_success,
                stop_after_tool_call=function_call.function.stop_after_tool_call or stop_execution_after_tool_call,
                metrics={"time": function_call_timer.elapsed},
            )

            # -*- Yield function call result
            yield ModelResponse(
                content=f"{function_call.get_call_str()} completed in {function_call_timer.elapsed:.4f}s.",
                tool_call=function_call_result.model_dump(
                    include={
                        "content",
                        "tool_call_id",
                        "tool_name",
                        "tool_args",
                        "tool_call_error",
                        "metrics",
                        "created_at",
                    }
                ),
                event=ModelResponseEvent.tool_call_completed.value,
            )

            # Add metrics to the model
            if "tool_call_times" not in self.metrics:
                self.metrics["tool_call_times"] = {}
            if function_call.function.name not in self.metrics["tool_call_times"]:
                self.metrics["tool_call_times"][function_call.function.name] = []
            self.metrics["tool_call_times"][function_call.function.name].append(function_call_timer.elapsed)

            # Add the function call result to the function call results
            function_call_results.append(function_call_result)
            if len(additional_messages_from_function_call) > 0:
                function_call_results.extend(additional_messages_from_function_call)
            self.function_call_stack.append(function_call)

            # -*- Check function call limit
            if self.tool_call_limit and len(self.function_call_stack) >= self.tool_call_limit:
                self.deactivate_function_calls()
                break  # Exit early if we reach the function call limit

    def handle_post_tool_call_messages(self, messages: List[Message], model_response: ModelResponse) -> ModelResponse:
        last_message = messages[-1]
        if last_message.stop_after_tool_call:
            logger.debug("Stopping execution as stop_after_tool_call=True")
            if (
                    last_message.role == "assistant"
                    and last_message.content is not None
                    and isinstance(last_message.content, str)
            ):
                if model_response.content is None:
                    model_response.content = ""
                model_response.content += last_message.content
        else:
            response_after_tool_calls = self.response(messages=messages)
            if response_after_tool_calls.content is not None:
                if model_response.content is None:
                    model_response.content = ""
                model_response.content += response_after_tool_calls.content
            if response_after_tool_calls.parsed is not None:
                # bubble up the parsed object, so that the final response has the parsed object
                # that is visible to the agent
                model_response.parsed = response_after_tool_calls.parsed
            if response_after_tool_calls.audio is not None:
                # bubble up the audio, so that the final response has the audio
                # that is visible to the agent
                model_response.audio = response_after_tool_calls.audio
        return model_response

    async def ahandle_post_tool_call_messages(
            self, messages: List[Message], model_response: ModelResponse
    ) -> ModelResponse:
        last_message = messages[-1]
        if last_message.stop_after_tool_call:
            logger.debug("Stopping execution as stop_after_tool_call=True")
            if (
                    last_message.role == "assistant"
                    and last_message.content is not None
                    and isinstance(last_message.content, str)
            ):
                if model_response.content is None:
                    model_response.content = ""
                model_response.content += last_message.content
        else:
            response_after_tool_calls = await self.aresponse(messages=messages)
            if response_after_tool_calls.content is not None:
                if model_response.content is None:
                    model_response.content = ""
                model_response.content += response_after_tool_calls.content
            if response_after_tool_calls.parsed is not None:
                # bubble up the parsed object, so that the final response has the parsed object
                # that is visible to the agent
                model_response.parsed = response_after_tool_calls.parsed
            if response_after_tool_calls.audio is not None:
                # bubble up the audio, so that the final response has the audio
                # that is visible to the agent
                model_response.audio = response_after_tool_calls.audio
        return model_response

    def handle_post_tool_call_messages_stream(self, messages: List[Message]) -> Iterator[ModelResponse]:
        last_message = messages[-1]
        if last_message.stop_after_tool_call:
            logger.debug("Stopping execution as stop_after_tool_call=True")
            if (
                    last_message.role == "assistant"
                    and last_message.content is not None
                    and isinstance(last_message.content, str)
            ):
                yield ModelResponse(content=last_message.content)
        else:
            yield from self.response_stream(messages=messages)

    async def ahandle_post_tool_call_messages_stream(self, messages: List[Message]) -> Any:
        last_message = messages[-1]
        if last_message.stop_after_tool_call:
            logger.debug("Stopping execution as stop_after_tool_call=True")
            if (
                    last_message.role == "assistant"
                    and last_message.content is not None
                    and isinstance(last_message.content, str)
            ):
                yield ModelResponse(content=last_message.content)
        else:
            async for model_response in self.aresponse_stream(messages=messages):  # type: ignore
                yield model_response

    def _process_string_image(self, image: str) -> Dict[str, Any]:
        """Process string-based image (base64, URL, or file path)."""

        # Process Base64 encoded image
        if image.startswith("data:image"):
            return {"type": "image_url", "image_url": {"url": image}}

        # Process URL image
        if image.startswith(("http://", "https://")):
            return {"type": "image_url", "image_url": {"url": image}}

        # Process local file image
        import base64
        import mimetypes
        from pathlib import Path

        path = Path(image)
        if not path.exists():
            raise FileNotFoundError(f"Image file not found: {image}")

        mime_type = mimetypes.guess_type(image)[0] or "image/jpeg"
        with open(path, "rb") as image_file:
            base64_image = base64.b64encode(image_file.read()).decode("utf-8")
            image_url = f"data:{mime_type};base64,{base64_image}"
            return {"type": "image_url", "image_url": {"url": image_url}}

    def _process_bytes_image(self, image: bytes) -> Dict[str, Any]:
        """Process bytes image data."""
        import base64

        base64_image = base64.b64encode(image).decode("utf-8")
        image_url = f"data:image/jpeg;base64,{base64_image}"
        return {"type": "image_url", "image_url": {"url": image_url}}

    def process_image(self, image: Any) -> Optional[Dict[str, Any]]:
        """Process an image based on the format."""

        if isinstance(image, dict):
            return {"type": "image_url", "image_url": image}

        if isinstance(image, str):
            return self._process_string_image(image)

        if isinstance(image, bytes):
            return self._process_bytes_image(image)

        logger.warning(f"Unsupported image type: {type(image)}")
        return None

    def add_images_to_message(self, message: Message, images: Optional[Sequence[Any]] = None) -> Message:
        """
        Add images to a message for the model. By default, we use the OpenAI image format but other Models
        can override this method to use a different image format.
        Args:
            message: The message for the Model
            images: Sequence of images in various formats:
                - str: base64 encoded image, URL, or file path
                - Dict: pre-formatted image data
                - bytes: raw image data

        Returns:
            Message content with images added in the format expected by the model
        """
        # If no images are provided, return the message as is
        if images is None or len(images) == 0:
            return message

        # Ignore non-string message content
        # because we assume that the images/audio are already added to the message
        if not isinstance(message.content, str):
            return message

        # Create a default message content with text
        message_content_with_image: List[Dict[str, Any]] = [{"type": "text", "text": message.content}]

        # Add images to the message content
        for image in images:
            try:
                image_data = self.process_image(image)
                if image_data:
                    message_content_with_image.append(image_data)
            except Exception as e:
                logger.error(f"Failed to process image: {str(e)}")
                continue

        # Update the message content with the images
        message.content = message_content_with_image
        return message

    def add_audio_to_message(self, message: Message, audio: Optional[Any] = None) -> Message:
        """
        Add audio to a message for the model. By default, we use the OpenAI audio format but other Models
        can override this method to use a different audio format.
        Args:
            message: The message for the Model
            audio: Pre-formatted audio data like {
                        "data": encoded_string,
                        "format": "wav"
                    }

        Returns:
            Message content with audio added in the format expected by the model
        """
        if audio is None:
            return message

        # If `id` is in the audio, this means the audio is already processed
        # This is used in multi-turn conversations
        if "id" in audio:
            message.content = ""
            message.audio = {"id": audio["id"]}
        # If `data` is in the audio, this means the audio is raw data
        # And an input audio
        elif "data" in audio:
            # Create a message with audio
            message.content = [
                {"type": "text", "text": message.content},
                {"type": "input_audio", "input_audio": audio},
            ]
        return message

    def get_system_message_for_model(self) -> Optional[str]:
        return self.system_prompt

    def get_instructions_for_model(self) -> Optional[List[str]]:
        return self.instructions

    def clear(self) -> None:
        """Clears the Model's state."""

        self.metrics = {}
        self.functions = None
        self.function_call_stack = None
        self.session_id = None

    def deep_copy(self, *, update: Optional[Dict[str, Any]] = None) -> "Model":
        new_model = self.model_copy(deep=True, update=update)
        # Clear the new model to remove any references to the old model
        new_model.clear()
        return new_model



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: 
"""
import os
import csv
import json
import ssl
import httpx
import numpy as np
from io import BytesIO
from pathlib import Path
from typing import Optional

from agentica.utils.log import logger


def read_json_file(path: Path) -> str:
    """Read JSON file."""
    try:
        if path.suffix == ".jsonl":
            with path.open("r", encoding="utf-8") as f:
                json_contents = [json.loads(line) for line in f]
        elif path.suffix == ".json":
            json_contents = json.loads(path.read_text("utf-8"))
        else:
            raise ValueError(f"Unsupported file format: {path.suffix}")
        if isinstance(json_contents, dict):
            json_contents = [json_contents]
        data = [json.dumps(content, ensure_ascii=False) for content in json_contents]
        return "\n".join(data)
    except Exception as e:
        logger.error(f"Error reading: {path}: {e}")
        return ""


def read_csv_file(path: Path, row_limit: Optional[int] = None) -> str:
    """Read CSV file."""
    try:
        with open(path, newline="") as csvfile:
            reader = csv.DictReader(csvfile)
            if row_limit is not None:
                csv_data = [row for row in reader][:row_limit]
            else:
                csv_data = [row for row in reader]
        return json.dumps(csv_data, ensure_ascii=False)
    except Exception as e:
        logger.error(f"Error reading: {path}: {e}")
        return ""


def read_txt_file(path: Path) -> str:
    """Read text file."""
    try:
        return path.read_text("utf-8")
    except Exception as e:
        logger.error(f"Error reading: {path}: {e}")
        return ""


def read_pdf_file(path: Path) -> str:
    """Read PDF file."""
    try:
        from docling.document_converter import DocumentConverter
    except (ImportError, ModuleNotFoundError):
        raise ImportError("`docling` not installed, please install using `pip install docling`")
    text = ""
    try:
        logger.info(f"Reading: {path}")
        converter = DocumentConverter()
        res = converter.convert(path)
        text = res.document.export_to_markdown()
    except Exception as e:
        logger.error(f"Error reading: {path}: {e}")
    return text


def read_pdf_url(url: str) -> str:
    """Read PDF file from URL."""
    text = ""
    try:
        from docling.document_converter import DocumentConverter
    except (ImportError, ModuleNotFoundError):
        raise ImportError("`docling` not installed, please install using `pip install docling`")
    try:
        logger.info(f"Reading: {url}")
        converter = DocumentConverter()
        res = converter.convert(url)
        text = res.document.export_to_markdown()
    except Exception as e:
        logger.error(f"Error reading: {url}: {e}")
    return text


def read_docx_file_by_docling(path: Path) -> str:
    """Read DOCX file using docling."""
    text = ""
    try:
        from docling.document_converter import DocumentConverter
    except (ImportError, ModuleNotFoundError):
        raise ImportError("`docling` not installed, please install using `pip install docling`")
    try:
        logger.info(f"Reading: {path}")
        converter = DocumentConverter()
        res = converter.convert(path)
        text = res.document.export_to_markdown()
    except Exception as e:
        logger.error(f"Error reading: {path}: {e}")
    return text


def read_docx_file(path: Path) -> str:
    """Read DOCX file."""
    try:
        from docx import Document as DocxDocument
        from docx.oxml.ns import qn
    except ImportError:
        raise ImportError("`python-docx` not installed, please install using `pip install python-docx`")

    def get_hyperlink_target(doc, r_id):
        """获取超链接目标URL或者路径."""
        if r_id in doc.part.rels:
            rel = doc.part.rels[r_id]
            if hasattr(rel, 'target_ref'):
                return rel.target_ref
        return None

    def is_image_url(url):
        image_extensions = {'.jpg', '.jpeg', '.png', '.gif', '.bmp', '.tiff', '.webp'}
        _, ext = os.path.splitext(url)
        return ext.lower() in image_extensions

    def get_numbering(paragraph):
        """获取段落的编号信息."""
        numbering = paragraph._element.xpath('.//w:numPr')
        if (numbering and len(numbering) > 0 and
                numbering[0].xpath('.//w:numId') and numbering[0].xpath('.//w:ilvl')):
            num_id = numbering[0].xpath('.//w:numId')[0].get(qn('w:val'))
            ilvl = numbering[0].xpath('.//w:ilvl')[0].get(qn('w:val'))
            return num_id, ilvl
        return None, None

    try:
        doc = DocxDocument(str(path))
        doc_content = ""
        numbering_dict = {}

        for paragraph in doc.paragraphs:
            para_text = ""
            num_id, ilvl = get_numbering(paragraph)
            if num_id and ilvl:
                if num_id not in numbering_dict:
                    numbering_dict[num_id] = {}
                if ilvl not in numbering_dict[num_id]:
                    numbering_dict[num_id][ilvl] = 1
                else:
                    numbering_dict[num_id][ilvl] += 1
                para_text += f"- "

            for run in paragraph.runs:
                run_text = run.text

                # 查找 Run 中的超链接
                hyperlink_elements = run._r.findall(".//w:hyperlink", namespaces=run._r.nsmap)
                for el in hyperlink_elements:
                    r_id = el.get(qn("r:id"))
                    if r_id:
                        hyperlink = get_hyperlink_target(doc, r_id)
                        if hyperlink:
                            text = "".join(node.text for node in el.findall(".//w:t", namespaces=run._r.nsmap))
                            md_url = f"![{text}]({hyperlink})" if is_image_url(hyperlink) else f"[{text}]({hyperlink})"
                            if text in run_text:
                                run_text = run_text.replace(text, md_url)
                            else:
                                run_text += md_url
                para_text += run_text

            # 查找段落中的超链接
            hyperlink_elements = paragraph._element.findall(".//w:hyperlink", namespaces=paragraph._element.nsmap)
            for el in hyperlink_elements:
                r_id = el.get(qn("r:id"))
                if r_id:
                    hyperlink = get_hyperlink_target(doc, r_id)
                    if hyperlink:
                        text = "".join(node.text for node in el.findall(".//w:t", namespaces=paragraph._element.nsmap))
                        md_url = f"![{text}]({hyperlink})" if is_image_url(hyperlink) else f"[{text}]({hyperlink})"
                        if text in para_text:
                            para_text = para_text.replace(text, md_url)
                        else:
                            para_text += md_url
            doc_content += para_text + "\n"

        # 处理插入的图片
        # for rel in doc.part.rels.values():
        #     if "image" in rel.target_ref and is_image_url(rel.target_ref):
        #         img_id = rel.rId
        #         img_part = doc.part.related_parts[img_id]
        #         img_data = img_part.blob
        #         # 生成保存图片的文件名
        #         os.makedirs(DATA_DIR, exist_ok=True)
        #         img_filename = os.path.join(DATA_DIR, f'image_{img_id}.png')
        #         # 保存图片数据到本地
        #         with open(img_filename, 'wb') as img_file:
        #             img_file.write(img_data)
        #         logger.debug(f"Image found: {rel.target_ref}, size: {len(img_data)} bytes")
        #         img_info = f"![Image]({img_filename})"
        #         doc_content += img_info + "\n"

        return doc_content
    except Exception as e:
        logger.error(f"Error reading: {path}: {e}")
        return ""


def read_excel_file(path: Path) -> str:
    """Read excel file."""
    text = ""
    try:
        from docling.document_converter import DocumentConverter
    except (ImportError, ModuleNotFoundError):
        raise ImportError("`docling` not installed, please install using `pip install docling`")
    try:
        logger.info(f"Reading: {path}")
        converter = DocumentConverter()
        res = converter.convert(path)
        text = res.document.export_to_markdown()
    except Exception as e:
        logger.error(f"Error reading: {path}: {e}")
    return text


if __name__ == '__main__':
    url = "https://arxiv.org/pdf/2412.15166"
    result = read_pdf_url(url)
    print(result)
